/// <reference types="node" />
/**
 * The keyword name shall be a left justified, eight-character,
 * space-filled, ASCII string with no embedded spaces. All digits
 * 0 through 9 (decimal ASCII codes 48 to 57, or hexadecimal 30
 * to 39) and upper case Latin alphabetic characters ’A’ through
 * ’Z’ (decimal 65 to 90 or hexadecimal 41 to 5A) are permitted;
 * lower-case characters shall not be used. The underscore (’ ’,
 * decimal 95 or hexadecimal 5F) and hyphen (’-’, decimal 45
 * or hexadecimal 2D) are also permitted. No other characters
 * are permitted.3 For indexed keyword names that have a single
 * positive integer index counter appended to the root name,
 * the counter shall not have leading zeros (e.g., NAXIS1, not
 * NAXIS001). Note that keyword names that begin with (or
 * consist solely of) any combination of hyphens, underscores, and
 * digits are legal.
 *
 * @param key Key word to test
 */
declare class MetaDataParser {
    static checkEntry(entry: string): void;
    static isValidKeyword(key: string): boolean;
    /**
     * If the two ASCII characters '= ' (decimal 61 followed
     * by decimal 32) are present in Bytes 9 and 10 of the keyword
     * record, this indicates that the keyword has a value field associated with it, unless it is one of the commentary keywords
     * defined in Sect. 4.4.2 (i.e., a HISTORY, COMMENT, or completely
     * blank keyword name), which, by definition, have no value.
     *
     * @param entry
     * @returns A trimmed keyword
     */
    static hasValue(entry: string): boolean;
    /**
     * Keyword name The first eight bytes of a keyword record, which
     * contain the ASCII name of a metadata quantity (unless it is blank).
     *
     * @param entry
     * @returns The trimmed keyword
     */
    static getKey(entry: string): string | undefined;
    static getValue(entry: string): string;
    /**
     * These keywords provide commentary information about the
     * contents or history of the FITS file and may occur any number of
     * times in a header. These keywords shall have no associated value
     * even if the value indicator characters `= ' appear in Bytes 9 and
     * 10 (hence it is recommended that these keywords not contain the
     * value indicator). Bytes 9 through 80 may contain any of the restricted set of ASCII-text characters, decimal 32 through 126
     * (hexadecimal 20 through 7E).
     * In earlier versions of this Standard continued string keywords (see Sect. 4.2.1.2) could be handled as commentary keywords if the relevant convention was not supported.
     * Now CONTINUE keywords shall be honoured as specified in
     * Sect. 4.2.1.2.
     *
     * @param key
     */
    static isCommentaryKeyword(key: string): boolean;
}
/**
 * FITS reader conforming to NASA's "Definition of the Flexible Image Transport System (FITS)"
 * https://fits.gsfc.nasa.gov/standard40/fits_standard40aa-le.pdf
 */
export declare class FitsReader {
    private static readonly HEADER_LENGTH;
    private static readonly HEADER_BLOCK_SIZE;
    private static getNumPixelBits;
    private static getMetadataValue;
    /**
     * BITPIX keyword. The value field shall contain an integer. The
     * absolute value is used in computing the sizes of data structures.
     * It shall specify the number of bits that represent a data value in
     * the associated data array. The only valid values of BITPIX are
     * given in Table 8. Writers of FITS arrays should select a BITPIX
     * data type appropriate to the form, range of values, and accuracy
     * of the data in the array
     *
     * Table 8: Interpretation of valid BITPIX value.
     * Value Data represented
     * 8 Character or unsigned binary integer
     * 16 16-bit two’s complement binary integer
     * 32 32-bit two’s complement binary integer
     * 64 64-bit two’s complement binary integer
     * −32 IEEE single-precision floating point
     * −64 IEEE double-precision floating point
     */
    private static bitsPerPixel;
    private static numAxis;
    private static axisSize;
    private static getDataSize;
    private static numBlocks;
    /**
     * A mapping of keyword names mapped to their values.
     */
    private readonly primaryHeaders;
    private readonly commentaryHeaders;
    private readonly extensionRawHeaders;
    private readonly extensionHeaders;
    private readonly extensionData;
    private readonly fileName;
    private rawData;
    private numBlocks;
    /**
     * Raw values of the primary header.
     */
    private rawHeaders;
    constructor(fileName: string, parser?: MetaDataParser);
    getRawData(): Promise<Buffer>;
    getNumBlocks(): Promise<number>;
    getDataBlock(blockNum: number): Promise<Buffer>;
    getMetadataBlock(blockNum: number): Promise<string[]>;
    /**
     * @returns An array of entries in the primary header as they appear in the file.
     */
    getRawPrimaryHeaders(): Promise<string[]>;
    /**
     * @returns A map of keys in the primary header mapped to their parsed values.
     */
    getPrimaryHeaders(): Promise<Map<string, string>>;
    /**
     * @returns True if this file conforms to NASA's format for FITS files (as indicated in metadata)
     */
    conformsToStandard(): Promise<boolean>;
    /**
     * Used in computing the sizes of data structures.
     * It shall specify the number of bits that represent a data value in
     * the associated data array
     */
    bitsPerPixel(): Promise<number>;
    /**
     * A non-negative integer no greater than 999 representing the
     * number of axes in the associated data array. A value of zero
     * signifies that no data follow the header in the HDU.
     */
    numAxis(): Promise<number>;
    /**
     * The NAXISn keywords must be present for
     * all values n = 1, . . . , NAXIS, in increasing order of n, and for
     * no other values of n. The value field of this indexed keyword
     * shall contain a non-negative integer representing the number of
     * elements along Axis n of a data array. A value of zero for any of
     * the NAXISn signifies that no data follow the header in the HDU
     * (however, the random-groups structure described in Sect. 6 has
     * NAXIS1 = 0, but will have data following the header if the other
     * NAXISn keywords are non-zero). If NAXIS is equal to 0, there
     * shall not be any NAXISn keywords.
     * @param axesIndex
     */
    sizeAxis(axesIndex: number): Promise<number>;
    /**
     * The total number of bits in the primary data array, exclusive
     * of fill that is needed after the data to complete the last 2880-byte
     * data block (Sect. 3.3.2), is given by the following expression:
     * Nbits = |BITPIX| × (NAXIS1 × NAXIS2 × · · · × NAXISm), (1)
     * where Nbits must be non-negative and is the number of bits excluding fill, m is the value of NAXIS, and BITPIX and the NAXISn
     * represent the values associated with those keywords. Note that
     * the random-groups convention in the primary array has a more complicated structure whose size is given by Eq. 4. The header
     * of the first FITS extension in the file, if present, shall start with
     * the first FITS block following the data block that contains the
     * last bit of the primary data array.
     */
    getDataSize(): Promise<number>;
    /**
     * The header
     * of the first FITS extension in the file, if present, shall start with
     * the first FITS block following the data block that contains the
     * last bit of the primary data array
     * @returns
     */
    firstExtensionBlockNum(): Promise<number>;
    getRawExtensions(): Promise<void>;
}
export {};
