"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.FitsReader = void 0;
const fs_jetpack_1 = require("fs-jetpack");
/**
 * The keyword name shall be a left justified, eight-character,
 * space-filled, ASCII string with no embedded spaces. All digits
 * 0 through 9 (decimal ASCII codes 48 to 57, or hexadecimal 30
 * to 39) and upper case Latin alphabetic characters ’A’ through
 * ’Z’ (decimal 65 to 90 or hexadecimal 41 to 5A) are permitted;
 * lower-case characters shall not be used. The underscore (’ ’,
 * decimal 95 or hexadecimal 5F) and hyphen (’-’, decimal 45
 * or hexadecimal 2D) are also permitted. No other characters
 * are permitted.3 For indexed keyword names that have a single
 * positive integer index counter appended to the root name,
 * the counter shall not have leading zeros (e.g., NAXIS1, not
 * NAXIS001). Note that keyword names that begin with (or
 * consist solely of) any combination of hyphens, underscores, and
 * digits are legal.
 *
 * @param key Key word to test
 */
class MetaDataParser {
    static checkEntry(entry) {
        if (entry.length !== 80) {
            throw new Error(`Metadata data entry: ${entry} is ${entry.length} characters. (Max 80)`);
        }
    }
    static isValidKeyword(key) {
        return key.charAt(0) !== " " && key.length === 8; // TODO other checks
    }
    /**
     * If the two ASCII characters '= ' (decimal 61 followed
     * by decimal 32) are present in Bytes 9 and 10 of the keyword
     * record, this indicates that the keyword has a value field associated with it, unless it is one of the commentary keywords
     * defined in Sect. 4.4.2 (i.e., a HISTORY, COMMENT, or completely
     * blank keyword name), which, by definition, have no value.
     *
     * @param entry
     * @returns A trimmed keyword
     */
    static hasValue(entry) {
        return entry.substring(8, 10) === "= ";
    }
    /**
     * Keyword name The first eight bytes of a keyword record, which
     * contain the ASCII name of a metadata quantity (unless it is blank).
     *
     * @param entry
     * @returns The trimmed keyword
     */
    static getKey(entry) {
        MetaDataParser.checkEntry(entry);
        const key = entry.substring(0, 8);
        if (MetaDataParser.isValidKeyword(key)) {
            return key.trim();
        }
    }
    static getValue(entry) {
        // TODO remove comments "/"
        MetaDataParser.checkEntry(entry);
        return entry.substring(9).split("/")[0].trim();
    }
    /**
     * These keywords provide commentary information about the
     * contents or history of the FITS file and may occur any number of
     * times in a header. These keywords shall have no associated value
     * even if the value indicator characters `= ' appear in Bytes 9 and
     * 10 (hence it is recommended that these keywords not contain the
     * value indicator). Bytes 9 through 80 may contain any of the restricted set of ASCII-text characters, decimal 32 through 126
     * (hexadecimal 20 through 7E).
     * In earlier versions of this Standard continued string keywords (see Sect. 4.2.1.2) could be handled as commentary keywords if the relevant convention was not supported.
     * Now CONTINUE keywords shall be honoured as specified in
     * Sect. 4.2.1.2.
     *
     * @param key
     */
    static isCommentaryKeyword(key) {
        return ["COMMENT", "HISTORY", "CONTINUE"].includes(key);
    }
}
/**
 * Table 8: Interpretation of valid BITPIX value.
 * Value Data represented
 * 8 Character or unsigned binary integer
 * 16 16-bit two’s complement binary integer
 * 32 32-bit two’s complement binary integer
 * 64 64-bit two’s complement binary integer
 * −32 IEEE single-precision floating point
 * −64 IEEE double-precision floating point
 */
var PixelDataType;
(function (PixelDataType) {
    PixelDataType["UnsignedBinary"] = "8";
    PixelDataType["Signed16"] = "16";
    PixelDataType["Signed32"] = "32";
    PixelDataType["Signed64"] = "64";
    PixelDataType["SingleFloat"] = "-32";
    PixelDataType["DoubleFloat"] = "-64";
})(PixelDataType || (PixelDataType = {}));
/**
 * FITS reader conforming to NASA's "Definition of the Flexible Image Transport System (FITS)"
 * https://fits.gsfc.nasa.gov/standard40/fits_standard40aa-le.pdf
 */
class FitsReader {
    constructor(fileName, parser = new MetaDataParser()) {
        /**
         * A mapping of keyword names mapped to their values.
         */
        this.primaryHeaders = new Map();
        // TODO need to be chronological
        this.commentaryHeaders = new Map();
        this.extensionRawHeaders = [];
        this.extensionHeaders = [];
        this.extensionData = [];
        this.fileName = fileName;
    }
    static getNumPixelBits(dataType) {
        if (Object.values(PixelDataType).includes(dataType)) {
            return Math.abs(parseInt(dataType));
        }
        else {
            throw new Error(`Invalid pixel data type ${dataType}`);
        }
    }
    static getMetadataValue(metadata, key) {
        const value = metadata.get(key);
        if (!value) {
            throw new Error(`Unable to find '${key} key in metadata`);
        }
        return value;
    }
    /**
     * BITPIX keyword. The value field shall contain an integer. The
     * absolute value is used in computing the sizes of data structures.
     * It shall specify the number of bits that represent a data value in
     * the associated data array. The only valid values of BITPIX are
     * given in Table 8. Writers of FITS arrays should select a BITPIX
     * data type appropriate to the form, range of values, and accuracy
     * of the data in the array
     *
     * Table 8: Interpretation of valid BITPIX value.
     * Value Data represented
     * 8 Character or unsigned binary integer
     * 16 16-bit two’s complement binary integer
     * 32 32-bit two’s complement binary integer
     * 64 64-bit two’s complement binary integer
     * −32 IEEE single-precision floating point
     * −64 IEEE double-precision floating point
     */
    static bitsPerPixel(metadata) {
        return FitsReader.getNumPixelBits(FitsReader.getMetadataValue(metadata, "BITPIX"));
    }
    static numAxis(metadata) {
        try {
            const naxis = FitsReader.getMetadataValue(metadata, "NAXIS");
            const num = parseInt(naxis);
            if (isNaN(num) || num < 0 || num > 999) {
                throw new Error(`Invalid number of axes: ${num}`);
            }
            return num;
        }
        catch (error) {
            console.error(error);
            throw new Error(`NAXIS keyword missing from metadata`);
        }
    }
    static axisSize(metadata, index) {
        const numAxis = FitsReader.numAxis(metadata);
        if (isNaN(numAxis) || numAxis <= 0) {
            throw new Error(`Cannot get index ${index} from data with ${numAxis} axis`);
        }
        const length = parseInt(FitsReader.getMetadataValue(metadata, `NAXIS${index}`));
        // TODO validate
        return length;
    }
    static getDataSize(metadata) {
        // TODO returned size is in bits not bytes
        const numAxis = FitsReader.numAxis(metadata);
        if (numAxis === 0) {
            return numAxis;
        }
        else {
            let prod = FitsReader.bitsPerPixel(metadata) / 8;
            for (let i = 1; i <= numAxis; i++) {
                const length = FitsReader.axisSize(metadata, i);
                // TODO check validity
                prod = prod * length;
            }
            return prod;
        }
    }
    static numBlocks(numBytes) {
        const div = numBytes / FitsReader.HEADER_BLOCK_SIZE;
        const num = Math.ceil(div);
        return num;
    }
    getRawData() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.rawData === undefined) {
                this.rawData = yield (0, fs_jetpack_1.readAsync)(this.fileName, "buffer");
            }
            if (this.rawData === undefined) {
                throw new Error(`Unable to read ${this.fileName}`);
            }
            return this.rawData;
        });
    }
    getNumBlocks() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.numBlocks) {
                this.numBlocks = FitsReader.numBlocks((yield this.getRawData()).length);
            }
            if (!this.numBlocks) {
                throw new Error(`Unable to get number of data blocks for ${this.fileName}`);
            }
            return this.numBlocks;
        });
    }
    getDataBlock(blockNum) {
        return __awaiter(this, void 0, void 0, function* () {
            return (yield this.getRawData()).subarray((blockNum - 1) * FitsReader.HEADER_BLOCK_SIZE, blockNum * FitsReader.HEADER_BLOCK_SIZE);
        });
    }
    getMetadataBlock(blockNum) {
        return __awaiter(this, void 0, void 0, function* () {
            const metadata = [];
            const buffer = yield this.getDataBlock(blockNum);
            for (let i = 0; i < FitsReader.HEADER_BLOCK_SIZE; i = i + FitsReader.HEADER_LENGTH) {
                metadata.push(buffer.subarray(i, i + FitsReader.HEADER_LENGTH).toString());
            }
            return metadata;
        });
    }
    /**
     * @returns An array of entries in the primary header as they appear in the file.
     */
    getRawPrimaryHeaders() {
        return __awaiter(this, void 0, void 0, function* () {
            // TODO validate headers (sec 4.4.1)
            if (!this.rawHeaders) {
                const metadata = [];
                let found = false;
                let blockNum = 1;
                while (!found) {
                    const block = yield this.getMetadataBlock(blockNum);
                    block.forEach((entry) => {
                        if (!found) {
                            found = entry.startsWith("END");
                        }
                        metadata.push(entry);
                        const key = MetaDataParser.getKey(entry);
                        if (key) {
                            if (MetaDataParser.isCommentaryKeyword(key)) {
                                if (key === "CONTINUE") {
                                    console.warn("Continue keyword not yet supported.");
                                }
                                if (!this.commentaryHeaders.get(key)) {
                                    this.commentaryHeaders.set(key, []);
                                }
                                const commentary = this.commentaryHeaders.get(key);
                                if (!commentary) {
                                    throw new Error("This really shouldn't happen");
                                }
                                commentary.push(entry.substring(8));
                            }
                            else {
                                const value = MetaDataParser.getValue(entry);
                                this.primaryHeaders.set(key, value);
                            }
                        }
                    });
                    blockNum++;
                }
                this.rawHeaders = metadata;
            }
            return this.rawHeaders;
        });
    }
    /**
     * @returns A map of keys in the primary header mapped to their parsed values.
     */
    getPrimaryHeaders() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.primaryHeaders.size === 0) {
                yield this.getRawPrimaryHeaders();
            }
            return this.primaryHeaders;
        });
    }
    /**
     * @returns True if this file conforms to NASA's format for FITS files (as indicated in metadata)
     */
    conformsToStandard() {
        return __awaiter(this, void 0, void 0, function* () {
            /**
             * 4.4.1.1
             * SIMPLE keyword. The value field shall contain a logical constant with the value T if the file conforms to this Standard. This
             * keyword is mandatory for the primary header and must not appear in extension headers.7 A value of F signifies that the file
             * does not conform to this Standard
             */
            const headers = yield this.getPrimaryHeaders();
            const simple = headers.get("SIMPLE");
            if (!simple) {
                throw new Error(`SIMPLE keyword missing from ${this.fileName} primary header metadata`);
            }
            return simple.startsWith("T");
        });
    }
    /**
     * Used in computing the sizes of data structures.
     * It shall specify the number of bits that represent a data value in
     * the associated data array
     */
    bitsPerPixel() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                return FitsReader.bitsPerPixel(yield this.getPrimaryHeaders());
            }
            catch (error) {
                console.error(error);
                throw new Error(`BITPIX keyword missing from ${this.fileName} primary header metadata`);
            }
        });
    }
    /**
     * A non-negative integer no greater than 999 representing the
     * number of axes in the associated data array. A value of zero
     * signifies that no data follow the header in the HDU.
     */
    numAxis() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                return FitsReader.numAxis(yield this.getPrimaryHeaders());
            }
            catch (error) {
                console.error(error);
                throw new Error(`${this.fileName} missing NAXIS metadata entry`);
            }
        });
    }
    /**
     * The NAXISn keywords must be present for
     * all values n = 1, . . . , NAXIS, in increasing order of n, and for
     * no other values of n. The value field of this indexed keyword
     * shall contain a non-negative integer representing the number of
     * elements along Axis n of a data array. A value of zero for any of
     * the NAXISn signifies that no data follow the header in the HDU
     * (however, the random-groups structure described in Sect. 6 has
     * NAXIS1 = 0, but will have data following the header if the other
     * NAXISn keywords are non-zero). If NAXIS is equal to 0, there
     * shall not be any NAXISn keywords.
     * @param axesIndex
     */
    sizeAxis(axesIndex) {
        return __awaiter(this, void 0, void 0, function* () {
            const headers = yield this.getPrimaryHeaders();
            const numAxis = yield this.numAxis();
            if (axesIndex === 0 || axesIndex > numAxis) {
                throw new Error(`Cannot get axis number ${axesIndex}. ${this.fileName} ${axesIndex === 0 ? "no data" : `has ${numAxis} Axis`}.`);
            }
            const key = `NAXIS${axesIndex}`;
            const naxisN = headers.get(key);
            if (!naxisN) {
                throw new Error(`NAXIS${axesIndex} keyword missing from ${this.fileName} primary header metadata`);
            }
            const size = parseInt(naxisN);
            if (isNaN(size) || size < 0) {
                throw new Error(`${size} is invalid for ${axesIndex} axis.`);
            }
            return size;
        });
    }
    /**
     * The total number of bits in the primary data array, exclusive
     * of fill that is needed after the data to complete the last 2880-byte
     * data block (Sect. 3.3.2), is given by the following expression:
     * Nbits = |BITPIX| × (NAXIS1 × NAXIS2 × · · · × NAXISm), (1)
     * where Nbits must be non-negative and is the number of bits excluding fill, m is the value of NAXIS, and BITPIX and the NAXISn
     * represent the values associated with those keywords. Note that
     * the random-groups convention in the primary array has a more complicated structure whose size is given by Eq. 4. The header
     * of the first FITS extension in the file, if present, shall start with
     * the first FITS block following the data block that contains the
     * last bit of the primary data array.
     */
    getDataSize() {
        return __awaiter(this, void 0, void 0, function* () {
            return FitsReader.getDataSize(yield this.getPrimaryHeaders());
        });
    }
    /**
     * The header
     * of the first FITS extension in the file, if present, shall start with
     * the first FITS block following the data block that contains the
     * last bit of the primary data array
     * @returns
     */
    firstExtensionBlockNum() {
        return __awaiter(this, void 0, void 0, function* () {
            const headers = yield this.getRawPrimaryHeaders();
            const headerBlocks = (headers.length * FitsReader.HEADER_LENGTH) /
                FitsReader.HEADER_BLOCK_SIZE;
            const dataSize = yield this.getDataSize();
            const dataBlocks = Math.ceil(dataSize / FitsReader.HEADER_BLOCK_SIZE);
            return headerBlocks + dataBlocks + 1;
        });
    }
    getRawExtensions() {
        return __awaiter(this, void 0, void 0, function* () {
            let blockNum = yield this.firstExtensionBlockNum();
            while (blockNum <= (yield this.getNumBlocks())) {
                const rawExtMetadata = [];
                const headers = new Map();
                const commentaries = new Map();
                let reachedEnd = false;
                while (!reachedEnd) {
                    const block = yield this.getMetadataBlock(blockNum);
                    block.forEach((entry) => {
                        if (!reachedEnd) {
                            reachedEnd = entry.startsWith("END");
                        }
                        rawExtMetadata.push(entry);
                        const key = MetaDataParser.getKey(entry);
                        if (key) {
                            if (MetaDataParser.isCommentaryKeyword(key)) {
                                if (key === "CONTINUE") {
                                    console.warn("Continue keyword not yet supported.");
                                }
                                if (!commentaries.get(key)) {
                                    commentaries.set(key, []);
                                }
                                const commentary = commentaries.get(key);
                                if (!commentary) {
                                    throw new Error("This really shouldn't happen");
                                }
                                commentary.push(entry.substring(8));
                            }
                            else {
                                const value = MetaDataParser.getValue(entry);
                                headers.set(key, value);
                            }
                        }
                    });
                    blockNum = blockNum + 1;
                }
                this.extensionRawHeaders.push(rawExtMetadata);
                this.extensionHeaders.push(headers);
                // TODO returned size is in bits not bytes
                const dataSize = FitsReader.getDataSize(headers);
                const numDataBlocks = FitsReader.numBlocks(dataSize);
                const dataBlocks = [];
                for (let i = 0; i < numDataBlocks; i++) {
                    const block = yield this.getDataBlock(blockNum);
                    dataBlocks.push(block);
                    blockNum = blockNum + 1;
                }
                // TODO raw data is in uint8 buffer. need to convert to bitpix?
                const rawData = Buffer.concat(dataBlocks);
                const parsed = [];
                let count = 0;
                // TODO only supporting 2D for now
                const numAxis = FitsReader.numAxis(headers);
                const bitpix = FitsReader.bitsPerPixel(headers);
                for (let y = 0; y < FitsReader.axisSize(headers, 2); y++) {
                    parsed[y] = [];
                    for (let x = 0; x < FitsReader.axisSize(headers, 1) && count * 4 < rawData.length; x++) {
                        const value = bitpix === 32
                            ? rawData.readFloatBE(count * 4)
                            : rawData.readInt8(count);
                        parsed[y][x] = value;
                        count++;
                    }
                }
                this.extensionData.push(parsed);
            }
            console.log();
        });
    }
}
exports.FitsReader = FitsReader;
// TODO prolly move to parser
FitsReader.HEADER_LENGTH = 80;
FitsReader.HEADER_BLOCK_SIZE = 2880;
